import git
import json
import logging
import os
import re
import requests
import tempfile

from enum import Enum
from requests.exceptions import HTTPError
from urllib.parse import urlsplit, urlunsplit, urljoin, quote


class RepoError(Exception):
    pass


class Visibility(Enum):
    """Gitlab API values for repo visibility"""
    private = 0
    internal = 10
    public = 20


class Access(Enum):
    guest = 10
    reporter = 20
    developer = 30
    master = 40
    owner = 50


class Repo(object):
    """Gitlab repo; manages API requests and various metadata"""

    PATH_RE = re.compile(
        r"^/(?P<namespace>[\w\-\.]+)/(?P<name>[\w\-\.]+)\.git$"
    )

    @classmethod
    def build_url(cls, url_base, namespace, name):
        """ Build a url for a repository """
        return url_base + "/" + namespace + "/" + name + ".git"

    @classmethod
    def from_url(cls, url, token):
        parts = urlsplit(url)
        if not parts.scheme:
            raise RepoError("{} is missing a scheme.".format(url))
        if not parts.netloc:
            raise RepoError("{} is missing a domain.".format(url))

        if parts.scheme != "https":
            logging.warning("Using scheme {} instead of https.", parts.scheme)

        match = cls.PATH_RE.match(parts.path)
        if not match:
            raise RepoError(
                "Bad path. Can't separate namespace " +
                "from repo name {}.".format(parts.path)
            )

        namespace = match.group("namespace")
        name = match.group("name")

        self = cls(urlunsplit((parts.scheme, parts.netloc, "", "", "")),
                   namespace, name, token, url)

        logging.debug(json.dumps(self.info))
        logging.debug("{} is valid.".format(self.name_with_namespace))

        return self

    @classmethod
    def _cls_gl_get(cls, url_base, path, token, params={}):
        """Make a Gitlab GET request"""
        params.update({"private_token": token})
        url = urljoin(url_base, "/api/v4" + path)
        r = requests.get(url, params=params)
        r.raise_for_status()
        return r.json()

    @classmethod
    def _cls_gl_post(cls, url_base, path, token, payload={}, params={}):
        """Make a Gitlab POST request"""
        params.update({"private_token": token})
        url = urljoin(url_base, "/api/v4" + path)
        r = requests.post(url, params=params, data=payload)
        r.raise_for_status()
        return r.json()

    @classmethod
    def _cls_gl_put(cls, url_base, path, token, payload={}, params={}):
        """Make a Gitlab PUT request"""
        params.update({"private_token": token})
        url = urljoin(url_base, "/api/v4" + path)
        r = requests.put(url, params=params, data=payload)
        r.raise_for_status()
        return r.json()

    @classmethod
    def _cls_gl_delete(cls, url_base, path, token, params={}):
        """Make a Gitlab DELETE request"""
        params.update({"private_token": token})
        url = urljoin(url_base, "/api/v4" + path)
        r = requests.delete(url, params=params)
        r.raise_for_status()
        return r.json()

    def __init__(self, url_base, namespace, name, token, url=None):
        self.url_base = url_base
        self.namespace = namespace
        self.name = name
        self.token = token

        if url is None:
            self.url = Repo.build_url(url_base, namespace, name)
        else:
            self.url = url

    @property
    def name_with_namespace(self):
        return "{}/{}".format(self.namespace, self.name)

    @property
    def namespace_id(self):
        if not self.already_exists():
            raise RepoError("Repo {} does not exist on Gitlab".format(self.name))

        return self.info["namespace"]["id"]

    @property
    def id(self):
        if not self.already_exists():
            raise RepoError("Repo {} does not exist on Gitlab".format(self.name))

        return self.info["id"]

    @property
    def info(self):
        if not hasattr(self, "_info"):
            quoted = quote(self.name_with_namespace, safe="")
            url = "/projects/{}".format(quoted)
            try:
                self._info = self._gl_get(url)
            except HTTPError as e:
                if e.response.status_code == 404:
                    logging.debug("Could not find repo with url {}/api/v4{}.".format(self.url_base,url))
                    self._info = None
                else:
                    raise
        return self._info

    @property
    def repo(self):
        if hasattr(self, "_repo"):
            return self._repo
        logging.debug("Repo not cloned yet!")
        return None

    @property
    def ssh_url(self):
        if not self.already_exists():
            raise RepoError("Repo {} does not exist on Gitlab".format(self.name))

        return self.info["ssh_url_to_repo"]

    def already_exists(self):
        if self.info:
            return True
        return False

    def clone_to(self, dir_name, branch=None):
        logging.debug("Cloning {}...".format(self.ssh_url))
        if branch:
            self._repo = git.Repo.clone_from(self.ssh_url, dir_name,
                                             branch=branch)
        else:
            self._repo = git.Repo.clone_from(self.ssh_url, dir_name)
        logging.info("Cloned {}.".format(self.name))
        return self._repo

    def delete(self):
        self._gl_delete("/projects/{}".format(self.id))
        logging.info("Deleted {}.".format(self.name))

    # TODO: this should really go elsewhere
    @classmethod
    def get_user_id(cls, username, url_base, token):
        data = cls._cls_gl_get(url_base, "/users", token,
                               params={"search": username})

        if len(data) == 0:
            logging.warn(
                "Did not find any users matching {}.".format(username)
            )
            raise RepoError("No user {}.".format(username))

        for result in data:
            if result["username"] == username:
                logging.info(
                    "Got id {} for user {}.".format(data[0]["id"], username)
                )
                return result["id"]

        # Fall back to first result if all else fails
        logging.warn("Got {} users for {}.".format(len(data), username))
        logging.warn("Failed to find an exact match for {}.".format(username))
        logging.info(
            "Got id {} for user {}.".format(data[0]["id"], data[0]["username"])
        )
        return data[0]["id"]

    def list_members(self):
        return self._gl_get(
            "/projects/{}/members".format(self.id)
        )

    def get_member(self, user_id):
        return self._gl_get(
            "/projects/{}/members/{}".format(self.id, user_id)
        )

    def add_member(self, user_id, level):
        payload = {
            "id": self.id,
            "user_id": user_id,
            "access_level": level.value
        }
        return self._gl_post("/projects/{}/members".format(self.id), payload)

    def edit_member(self, user_id, level):
        payload = {
            "id": self.id,
            "user_id": user_id,
            "access_level": level.value
        }
        return self._gl_put(
            "/projects/{}/members/{}".format(self.id, user_id), payload
        )

    def delete_member(self, user_id):
        return self._gl_delete(
            "/projects/{}/members/{}".format(self.id, user_id)
        )

    def list_commits(self, ref_name="master"):
        params = {
            "id": self.id,
            "ref_name": ref_name
        }
        return self._gl_get(
            "/projects/{}/repository/commits".format(self.id), params
        )

    def list_branches(self):
        return self._gl_get(
            "/projects/{}/repository/branches".format(self.id)
        )

    def get_branch(self, branch):
        return self._gl_get(
            "/projects/{}/repository/branches/{}".format(self.id, branch)
        )

    def archive(self):
        return self._gl_post("/projects/{}/archive".format(self.id))

    def unarchive(self):
        return self._gl_post("/projects/{}/unarchive".format(self.id))

    def protect(self, branch="master", developer_push=True, developer_merge=True): # NOTE: these are not the same defaults that Gitlab uses
        params = {
            "developers_can_push": developer_push,
            "developers_can_merge": developer_merge,
          }
        return self._gl_put("/projects/{}/repository/branches/{}/protect".format(self.id, branch), params)

    def unprotect(self, branch="master"):
        return self._gl_put("/projects/{}/repository/branches/{}/unprotect".format(self.id, branch))

    def _gl_get(self, path, params={}):
        return self.__class__._cls_gl_get(
            self.url_base, path, self.token, params
        )

    def _gl_post(self, path, payload={}, params={}):
        return self.__class__._cls_gl_post(
            self.url_base, path, self.token, payload
        )

    def _gl_put(self, path, payload={}, params={}):
        return self.__class__._cls_gl_put(
            self.url_base, path, self.token, payload
        )

    def _gl_delete(self, path, params={}):
        return self.__class__._cls_gl_delete(
            self.url_base, path, self.token, params
        )


class BaseRepo(Repo):

    @classmethod
    def new(cls, name, namespace, url_base, token):
        namespaces = cls._cls_gl_get(url_base, "/namespaces",
                                     token, {"search": namespace})
        logging.debug(
            "Got {} namespaces matching {}.".format(len(namespaces), namespace)
        )
        logging.debug(
            "Using namespace " +
            "{} with ID {}.".format(namespaces[0]["path"], namespaces[0]["id"])
        )

        payload = {
            "name": name,
            "namespace_id": namespaces[0]["id"],
            "issues_enabled": False,
            "merge_requests_enabled": False,
            "builds_enabled": False,
            "wiki_enabled": False,
            "snippets_enabled": True,  # Why not?
            "visibility_level": Visibility.private.value,
        }

        result = cls._cls_gl_post(url_base, "/projects", token, payload)

        return cls.from_url(result["http_url_to_repo"], token)

    def push_to(self, student_repo, branch="master"):
        r = git.Remote.add(self.repo, student_repo.name,
                           student_repo.ssh_url)
        r.push(branch)
        logging.info("Pushed {} to {}.".format(self.name, student_repo.name))


class StudentRepo(Repo):
    """Repository for a student's solution to a homework assignment"""

    @classmethod
    def new(cls, base_repo, semester, section, username, token):
        """Create a new repository on GitLab"""
        payload = {
            "name": cls.name(semester, section, base_repo.name, username),
            "namespace_id": base_repo.namespace_id,
            "issues_enabled": False,
            "merge_requests_enabled": False,
            "builds_enabled": False,
            "wiki_enabled": False,
            "snippets_enabled": True,  # Why not?
            "visibility_level": Visibility.private.value,
        }

        result = cls._cls_gl_post(base_repo.url_base, "/projects",
                                  token, payload)

        return cls.from_url(result["http_url_to_repo"], token)

    @classmethod
    def name(cls, semester, section, assignment, user):
        fmt = {
            "semester": semester,
            "section": section,
            "assignment": assignment,
            # Replace .s with -s
            "user": user.translate(str.maketrans(".", "-"))
        }

        return "{semester}-{section}-{assignment}-{user}".format(**fmt)

    def push(self, base_repo, branch="master"):
        """Push base_repo code to this repo"""
        base_repo.push_to(self, branch)


if __name__ == "__main__":
    # Need to make this module if'n you're going to run this
    from secret import token

    logging.basicConfig(level=logging.INFO)

    b = BaseRepo("https://git.mst.edu/2016-Spring-CS-2001/hw01.git", token)
    print(json.dumps(b.info, indent=8))
    with tempfile.TemporaryDirectory() as tmpdirname:
        b.clone_to(tmpdirname)
        print(os.listdir(tmpdirname))

    StudentRepo.new(b, "2016SP", "A", "mwwcp2", token)



import logging
import datetime

from assigner.config import config_context

help="Interactively initialize a new configuration"

logger = logging.getLogger(__name__)

def prompt(explanation, default=None):
    prompt_string = ''
    if default is not None:
        prompt_string = "{} (default: {}): ".format(explanation, default)
    else:
        prompt_string = "{}: ".format(explanation)
    value = input(prompt_string)

    if value == '':
        if default is not None:
            return default

        return prompt(explanation, default)

    return value

def guess_semester():
    now = datetime.datetime.now()
    if now.month < 5:
        semester = 'SP'
    elif now.month < 8:
        semester = 'SS'
    else:
        semester = 'FS'

    return "{}-{}".format(now.year, semester)


@config_context
def init(conf, args):
    conf['gitlab-host'] = "https://{}".format(prompt("Gitlab server to use", "gitlab.com"))
    conf['token'] = prompt("Gitlab access token (from {}/profile/personal_access_tokens)".format(conf['gitlab-host']))
    conf['semester'] = prompt("Year and semester, in the format YYYY-(FS|SP|SS)", guess_semester())
    conf['namespace'] = prompt("Gitlab group to create repositories under", "{}-CS1001".format(conf['semester']))
    do_canvas = input("Do you want to configure Canvas integration? [y/N]: ")
    if do_canvas.lower() == 'y':
        conf['canvas-host'] = prompt("Canvas server to use (???.instructure.com)")
        conf['canvas-token'] = prompt("Canvas access token (from {}/profile/settings)".format(conf['canvas-host']))

    print("Congratulations, you're ready to go!")

def setup_parser(parser):
    parser.set_defaults(run=init)


import logging

from assigner.config import config_context, DuplicateUserError
from prettytable import PrettyTable
from assigner.roster_util import get_filtered_roster, add_to_roster

help = "Manage class roster"

logger = logging.getLogger(__name__)

@config_context
def list_students(conf, args):
    """List students in the roster
    """
    output = PrettyTable(["#", "Name", "Username", "Section"])
    for idx,student in enumerate(get_filtered_roster(conf.roster, args.section, None)):
        output.add_row((idx+1, student["name"], student["username"], student["section"]))

    print(output)

@config_context
def add_student(conf, args):
    """Add a student to the roster
    """
    try:
        add_to_roster(conf, conf.roster, args.name, args.username, args.section, args.force)
    except DuplicateUserError:
        logger.error("Student already exists in roster!")

@config_context
def remove_student(conf, args):
    """Remove a student from the roster
    """
    previous_len = len(conf.roster)

    idxs = []
    for idx, student in enumerate(conf.roster):
        if student['username'] == args.username:
            idxs.append(idx)

    offset = 0
    for idx in idxs:
        del conf.roster[idx - offset]
        offset += 1

    logger.info("Removed {} entries from the roster".format(previous_len - len(conf.roster)))

def setup_parser(parser):
    subparsers = parser.add_subparsers(title='Canvas commands')

    list_parser = subparsers.add_parser('list', help='Print the roster')
    list_parser.add_argument("--section", help="Section to list")
    list_parser.set_defaults(run=list_students)

    add_parser = subparsers.add_parser('add', help='Add a student to the roster')
    add_parser.add_argument("name", help="Name of student")
    add_parser.add_argument("username", help="Username of student")
    add_parser.add_argument("section", help="Section of student")
    add_parser.add_argument("--force", action="store_true", help="Add duplicate student anyway")
    add_parser.set_defaults(run=add_student)


    remove_parser = subparsers.add_parser('remove', help='Remove a student from the roster')
    remove_parser.add_argument("username", help="Username of student to remove")
    remove_parser.set_defaults(run=remove_student)

    help_parser = subparsers.add_parser("help",
                                        help="Show this help screen and exit")
    help_parser.set_defaults(run=lambda x: parser.print_help())


import logging
import tempfile
import time

from assigner.roster_util import get_filtered_roster
from assigner.baserepo import BaseRepo, StudentRepo
from assigner.config import config_context
from requests.exceptions import HTTPError

help="Assign a base repo to students"

logger = logging.getLogger(__name__)


@config_context
def assign(conf, args):
    """Creates homework repositories for an assignment for each student
    in the roster.
    """
    hw_name = args.name
    if args.branch:
        branch = args.branch
    else:
        branch = "master"
    dry_run = args.dry_run
    force = args.force
    host = conf.gitlab_host
    namespace = conf.namespace
    token = conf.token
    semester = conf.semester

    roster = get_filtered_roster(conf.roster, args.section, args.student)

    actual_count = 0  # Represents the number of repos actually pushed to
    student_count = len(roster)

    with tempfile.TemporaryDirectory() as tmpdirname:
        print("Assigning '{}' to {} student{} in {}.".format(
            hw_name, student_count,
            "s" if student_count != 1 else "",
            "section " + args.section if args.section else "all sections")
        )
        base = BaseRepo(host, namespace, hw_name, token)
        if not dry_run:
            base.clone_to(tmpdirname, branch)
        if force:
            logging.warning("Repos will be overwritten.")
        for i, student in enumerate(roster):
            username = student["username"]
            student_section = student["section"]
            full_name = StudentRepo.name(semester, student_section,
                                         hw_name, username)
            repo = StudentRepo(host, namespace, full_name, token)

            print("{}/{} - {}".format(i+1, student_count, full_name))
            if not repo.already_exists():
                if not dry_run:
                    repo = StudentRepo.new(base, semester, student_section,
                                           username, token)
                    repo.push(base, branch)
                    repo.protect(branch)
                actual_count += 1
                logging.debug("Assigned.")
            elif force:
                logging.info("{}: Already exists.".format(full_name))
                logging.info("{}: Deleting...".format(full_name))
                if not dry_run:
                    repo.delete()

                    # Gitlab will throw a 400 if you delete and immediately recreate a repo.
                    # We retry w/ exponential backoff up to 5 times
                    wait = 0.1
                    retries = 0
                    while True:
                        try:
                            repo = StudentRepo.new(base, semester, student_section,
                                                   username, token)
                            logger.debug("Success!")
                            break
                        except HTTPError as e:
                            if retries >= 5 or e.response.status_code != 400:
                                logger.debug("Critical Failure!")
                                raise
                            logger.debug("Failed, retrying...")

                        # Delay and try again
                        time.sleep(wait * 2**retries)
                        retries += 1

                    repo.push(base, branch)
                    repo.protect(branch)
                actual_count += 1
                logging.debug("Assigned.")
            elif args.branch:
                logging.info("{}: Already exists.".format(full_name))
                # If we have an explicit branch, push anyways
                if not dry_run:
                    repo.push(base, branch)
                    repo.protect(branch)
                actual_count += 1
                logging.debug("Assigned.")
            else:
                logging.warning("Skipping...")
            i += 1

    print("Assigned '{}' to {} student{}.".format(
        hw_name,
        actual_count,
        "s" if actual_count != 1 else ""
    ))
    if actual_count == 0:
        logging.warning(
            "Consider using --force if you want to override existing repos."
        )


def setup_parser(parser):
    parser.add_argument("name",
                           help="Name of the assignment to assign.")
    parser.add_argument("--branch", nargs="?",
                           help="Branch to push")
    parser.add_argument("--section", nargs="?",
                           help="Section to assign homework to")
    parser.add_argument("--student", metavar="id",
                           help="ID of the student to assign to.")
    parser.add_argument("--dry-run", action="store_true",
                           help="Don't actually do it.")
    parser.add_argument("-f", "--force", action="store_true", dest="force",
                           help="Delete and recreate already existing " +
                                "student repos.")
    parser.set_defaults(run=assign)



import http.client
import json
import logging
import time

__author__ = 'Islam Elnabarawy'


class CanvasAPI:
    REQUEST_HEADER = {
        "Authorization": "Bearer ",
        "Content-Type": "application/json"
    }

    def __init__(self, canvas_token, website_root):
        self.REQUEST_HEADER['Authorization'] += canvas_token
        self.website_root = website_root

    def _request(self, method: str, url: str, params: dict = None, retries: int = 3) -> http.client.HTTPResponse:
        tries = 0
        while tries <= retries:
            try:
                connection = http.client.HTTPSConnection(self.website_root, 443)
                connection.connect()
                header = self.REQUEST_HEADER
                connection.request(method, url, (json.dumps(params) if params is not None else None), header)
                return connection.getresponse()
            except Exception as ex:
                tries += 1
                if tries > retries:
                    raise ex
                logging.warning("Caught exception in request after %d tries. Will retry %d more times.",
                                tries, retries - tries, exc_info=True)
                time.sleep(1)

    def _decode_links(self, link):
        """
        Extract pagination links from the Canvas API response header's Links field
        :param link: value of the Links field in Canvas API's response header
        :return: A dictionary with the name and value of each link
        """
        links = link.split(',')
        result = {}
        for l in links:
            f, s = l.split(';')
            result[s[6:-1]] = f[1:-1].replace(self.website_root, '')
        return result

    def _get_all_pages(self, url: str, params: dict = None) -> list:
        """
        Get the full results from a query by following pagination links until the last page
        :param url: The URL for the query request
        :param params: A dictionary of parameter names and values to be sent with the request
        :return: The full list of result objects returned by the query
        """
        response = self._request('GET', url, params)
        result = json.loads(response.read().decode())
        links = self._decode_links(response.getheader("Link"))
        count = len(result)
        page = 1
        while 'next' in links:
            logging.debug("Got links:\n" + json.dumps(links, sort_keys=True, indent=2))
            page += 1
            next_url = url.split('?')[0] + '?page=%s&per_page=%s' % (page, count)
            logging.info("Getting next page for " + url + " via " + next_url)
            response = self._request('GET', next_url, params)
            result.extend(json.loads(response.read().decode()))
            links = self._decode_links(response.getheader("Link"))

        return result

    def get_instructor_courses(self):
        get = lambda x: self._get_all_pages('/api/v1/courses',
                                     {'enrollment_type': x, 'state': ['available']})
        result = get('teacher')
        result.extend(get('ta'))
        result.extend(get('grader'))
        return result

    def get_course_students(self, course_id):
        result = self._get_all_pages('/api/v1/courses/%s/users?per_page=50' % course_id,
                                     {'enrollment_type': ['student'], 'enrollment_state': ['active']})
        return result



import csv
import logging
import re

from assigner.config import config_context, DuplicateUserError
from assigner.roster_util import add_to_roster

help="Import users from a csv"

logger = logging.getLogger(__name__)


@config_context
def import_students(conf, args):
    """Imports students from a CSV file to the roster.
    """
    section = args.section

    # TODO: This should probably move to another file
    email_re = re.compile(r"^(?P<user>[^@]+)")
    with open(args.file) as fh:
        reader = csv.reader(fh)

        if "roster" not in conf:
            conf["roster"] = []

        # Note: This is incredibly hardcoded.
        # However, peoplesoft never updates anything, so we're probably good.
        reader.__next__()  # Skip the header
        count = 0
        for row in reader:
            count += 1
            match = email_re.match(row[4])

            try:
                add_to_roster(conf, conf.roster, row[3], match.group("user"), section, args.force)
            except DuplicateUserError:
                logger.warning("User {} is already in the roster, skipping".format(match.group("user")))

    print("Imported {} students.".format(count))

def setup_parser(parser):
    parser.add_argument("file", help="CSV file to import from")
    parser.add_argument("section", help="Section being imported")
    parser.add_argument("--force", action="store_true", help="Import duplicate students anyway")
    parser.set_defaults(run=import_students)



import logging

from assigner.canvas import CanvasAPI
from assigner.config import config_context,DuplicateUserError

from prettytable import PrettyTable

from assigner.roster_util import add_to_roster

help = "Get Canvas course information"

logger = logging.getLogger(__name__)

@config_context
def import_from_canvas(conf, args):
    """Imports students from a Canvas course to the roster.
    """
    if 'canvas-token' not in conf:
        logger.error("canvas-token configuration is missing! Please set the Canvas API access "
                      "token before attempting to import users from Canvas")
        print("Import from canvas failed: missing Canvas API access token.")
        return

    if "roster" not in conf:
        conf["roster"] = []

    course_id = args.id
    section = args.section
    force = args.force

    canvas = CanvasAPI(conf["canvas-token"], conf["canvas-host"])

    students = canvas.get_course_students(course_id)

    for s in students:
        if 'sis_user_id' not in s:
            logger.error("Could not get username for {}".format(s['sortable_name']))

        try:
            add_to_roster(conf, conf.roster, s['sortable_name'], s['sis_user_id'], section, force)
        except DuplicateUserError:
            logger.warning("User {} is already in the roster, skipping".format(s['sis_user_id']))

    print("Imported {} students.".format(len(students)))

@config_context
def print_canvas_courses(conf, args):
    """Show a list of current teacher's courses from Canvas via the API.
    """
    if 'canvas-token' not in conf:
        logger.error("canvas-token configuration is missing! Please set the Canvas API access "
                      "token before attempting to use Canvas API functionality")
        print("Canvas course listing failed: missing Canvas API access token.")
        return

    canvas = CanvasAPI(conf["canvas-token"], conf["canvas-host"])

    courses = canvas.get_instructor_courses()

    if not courses:
        print("No courses found where current user is a teacher.")
        return

    output = PrettyTable(["#", "ID", "Name"])
    output.align["Name"] = "l"

    for ix, c in enumerate(sorted(courses, key=lambda c: c['id'], reverse=True)):
        output.add_row((ix+1, c['id'], c['name']))

    print(output)

def setup_parser(parser):
    subparsers = parser.add_subparsers(title='Canvas commands')

    list_parser = subparsers.add_parser('list', help='List Canvas courses where you are a teacher, TA, or grader')
    list_parser.set_defaults(run=print_canvas_courses)

    import_parser = subparsers.add_parser('import', help='Import the roster from a specific Canvas course')
    import_parser.add_argument("id", help="Canvas ID for course to import from")
    import_parser.add_argument("section", help="Section being imported")
    import_parser.add_argument("--force", action="store_true", help="Import duplicate students anyway")
    import_parser.set_defaults(run=import_from_canvas)

    help_parser = subparsers.add_parser("help",
                                        help="Show this help screen and exit")
    help_parser.set_defaults(run=lambda x: parser.print_help())



import logging

from assigner import manage_repos

help="Unprotect a repo branch"

logger = logging.getLogger(__name__)

def unprotect(args):
    """Unprotect a branch in each student's repository so they can force push to it."""
    branch = args.branch if args.branch else "master"
    return manage_repos(args, lambda repo: repo.unprotect(branch))

def setup_parser(parser):
    parser.add_argument("name",
                           help="Name of the assignment to unprotect.")
    parser.add_argument("--branch",
                           help="Branch to unprotect (default: master).")
    parser.add_argument("--section", nargs="?",
                           help="Section to unprotect")
    parser.add_argument("--student", metavar="id",
                           help="ID of student whose assignment to unprotect.")
    parser.add_argument("--dry-run", action="store_true",
                           help="Don't actually do it.")
    parser.set_defaults(run=unprotect)




import logging

from assigner import manage_users
from assigner.baserepo import Access

help="Lock students out of repos"

logger = logging.getLogger(__name__)


def lock(args):
    """Sets each student to Reporter status on their homework repository so
    they cannot push changes, etc.
    """
    return manage_users(args, Access.reporter)


def setup_parser(parser):
    parser.add_argument("name",
                           help="Name of the assignment to lock.")
    parser.add_argument("--section", nargs="?",
                           help="Section to lock")
    parser.add_argument("--student", metavar="id",
                           help="ID of student whose assignment needs " +
                                "locking.")
    parser.add_argument("--dry-run", action="store_true",
                           help="Don't actually do it.")
    parser.set_defaults(run=lock)



import logging

from assigner import manage_repos

help="Protect a repo branch"

logger = logging.getLogger(__name__)


def protect(args):
    """Protect a branch in each student's repository so they cannot force push to it."""
    branch = args.branch if args.branch else "master"
    return manage_repos(args, lambda repo: repo.protect(branch))

def setup_parser(parser):
    parser.add_argument("name",
                           help="Name of the assignment to protect.")
    parser.add_argument("--branch",
                           help="Branch to protect (default: master).")
    parser.add_argument("--section", nargs="?",
                           help="Section to protect")
    parser.add_argument("--student", metavar="id",
                           help="ID of student whose assignment to protect.")
    parser.add_argument("--dry-run", action="store_true",
                           help="Don't actually do it.")
    parser.set_defaults(run=protect)



import logging

from assigner import manage_repos

help="Archive repos"

logger = logging.getLogger(__name__)


def archive(args):
    """Archive each student repository so it won't show up in the project list.
    """
    return manage_repos(args, lambda repo: repo.archive())


def setup_parser(parser):
    parser.add_argument("name",
                           help="Name of the assignment to archive.")
    parser.add_argument("--section", nargs="?",
                           help="Section to archive")
    parser.add_argument("--student", metavar="id",
                           help="ID of student whose assignment to archive.")
    parser.add_argument("--dry-run", action="store_true",
                           help="Don't actually do it.")
    parser.set_defaults(run=archive)



import logging
from datetime import datetime

from requests.exceptions import HTTPError
from prettytable import PrettyTable
from progressbar import ProgressBar

from assigner.roster_util import get_filtered_roster
from assigner.config import config_context
from assigner.baserepo import Access, Repo, StudentRepo, RepoError

help="Retrieve status of repos"

logger = logging.getLogger(__name__)


@config_context
def status(conf, args):
    """Retrieves and prints the status of repos"""
    hw_name = args.name

    if not hw_name:
        raise ValueError("Missing assignment name.")

    host = conf.gitlab_host
    namespace = conf.namespace
    token = conf.token
    semester = conf.semester

    roster = get_filtered_roster(conf.roster, args.section, args.student)
    sort_key = args.sort

    if sort_key:
        roster.sort(key=lambda s: s[sort_key])

    output = PrettyTable([
        "#", "Sec", "SID", "Name", "Status", "Branches",
        "HEAD", "Last Commit Author", "Last Commit Time"])
    output.align["Name"] = "l"
    output.align["Last Commit Author"] = "l"

    progress = ProgressBar(max_value=len(roster))

    for i, student in enumerate(roster):
        progress.update(i)

        name = student["name"]
        username = student["username"]
        student_section = student["section"]
        full_name = StudentRepo.name(semester, student_section,
                                     hw_name, username)

        row = [i+1, student_section, username, name, "", "", "", "", ""]

        try:
            repo = StudentRepo(host, namespace, full_name, token)

            if not repo.already_exists():
                row[4] = "Not Assigned"
                output.add_row(row)
                continue

            if "id" not in student:
                try:
                    student["id"] = Repo.get_user_id(username, host, token)
                except RepoError:
                    row[4] = "No Gitlab user"
                    output.add_row(row)
                    continue

            members = repo.list_members()
            if student["id"] not in [s["id"] for s in members]:
                row[4] = "Not Opened"
                output.add_row(row)
                continue

            if repo.info["archived"]:
                row[4] = 'Archived'
            else:
                level = Access([s["access_level"] for s in members if s["id"] == student["id"]][0])
                row[4] = "Open" if level is Access.developer else "Locked"

            branches = repo.list_branches()

            if branches:
                row[5] = ", ".join([b["name"] for b in branches])

            commits = repo.list_commits()

            if commits:
                head = commits[0]
                row[6] = head["short_id"]
                row[7] = head["author_name"]
                created_at = head["created_at"]
                # Fix UTC offset format in GitLab's datetime
                created_at = created_at[:-6] + created_at[-6:].replace(':', '')
                row[8] = datetime.strptime(created_at, "%Y-%m-%dT%H:%M:%S.%f%z")

            output.add_row(row)

        except HTTPError:
            raise

    progress.finish()
    print(output)


def setup_parser(parser):
    parser.add_argument("--section", nargs="?",
                           help="Section to get status of")
    parser.add_argument("--student", metavar="id",
                           help="ID of student.")
    parser.add_argument("--sort", nargs="?", default="name",
                           choices=["name", "username"],
                           help="Key to sort users by.")
    parser.add_argument("name", nargs="?",
                           help="Name of the assignment to look up.")
    parser.set_defaults(run=status)



import logging

from requests.exceptions import HTTPError

from assigner.roster_util import get_filtered_roster
from assigner.baserepo import Access, Repo, RepoError, StudentRepo
from assigner.config import config_context

help="Grants students access to their repos"

logger = logging.getLogger(__name__)


@config_context
def open_assignment(conf, args):
    """Adds each student in the roster to their respective homework
    repositories as Developers so they can pull/commit/push their work.
    """
    hw_name = args.name
    host = conf.gitlab_host
    namespace = conf.namespace
    token = conf.token
    semester = conf.semester

    roster = get_filtered_roster(conf.roster, args.section, args.student)

    count = 0
    for student in roster:
        username = student["username"]
        student_section = student["section"]
        full_name = StudentRepo.name(semester, student_section,
                                     hw_name, username)

        try:
            repo = StudentRepo(host, namespace, full_name, token)
            if "id" not in student:
                student["id"] = Repo.get_user_id(username, host, token)

            repo.add_member(student["id"], Access.developer)
            count += 1
        except RepoError:
            logging.warn("Could not add {} to {}.".format(username, full_name))
        except HTTPError:
            raise

    print("Granted access to {} repositories.".format(count))


def setup_parser(parser):
    parser.add_argument("name", help="Name of the assignment to grant " +
                                        "access to")
    parser.add_argument("--section", nargs="?",
                           help="Section to grant access to")
    parser.add_argument("--student", metavar="id",
                           help="ID of the student to assign to.")
    parser.set_defaults(run=open_assignment)



import logging
import os

from requests.exceptions import HTTPError

from assigner.roster_util import get_filtered_roster
from assigner.baserepo import RepoError, StudentRepo
from assigner.config import config_context

help="Clone student repos"

logger = logging.getLogger(__name__)


@config_context
def get(conf, args):
    """Creates a folder for the assignment in the CWD (or <path>, if specified)
    and clones each students' repository into subfolders.
    """
    hw_name = args.name
    hw_path = args.path
    host = conf.gitlab_host
    namespace = conf.namespace
    token = conf.token
    semester = conf.semester

    roster = get_filtered_roster(conf.roster, args.section, args.student)

    path = os.path.join(hw_path, hw_name)
    os.makedirs(path, mode=0o700, exist_ok=True)

    count = 0
    for student in roster:
        username = student["username"]
        student_section = student["section"]
        full_name = StudentRepo.name(semester, student_section,
                                     hw_name, username)

        try:
            repo = StudentRepo(host, namespace, full_name, token)
            repo.clone_to(os.path.join(path, username))
            count += 1
        except RepoError as e:
            logging.warn(str(e))
        except HTTPError as e:
            if e.response.status_code == 404:
                logging.warn("Repository {} does not exist.".format(full_name))
            else:
                raise

    print("Cloned {} repositories.".format(count))


def setup_parser(parser):
    parser.add_argument("name",
                           help="Name of the assignment to retrieve.")
    parser.add_argument("path", default=".", nargs="?",
                           help="Path to clone student repositories to")
    parser.add_argument("--section", nargs="?",
                           help="Section to retrieve")
    parser.add_argument("--student", metavar="id",
                           help="ID of student whose assignment needs " +
                                "retrieving.")
    parser.set_defaults(run=get)



from assigner.config import DuplicateUserError
from assigner.baserepo import Repo, RepoError

import logging
logger = logging.getLogger(__name__)

def get_filtered_roster(roster, section, target):
    if target:
        roster = [s for s in roster if s["username"] == target]
    elif section:
        roster = [s for s in roster if s["section"] == section]
    if not roster:
        raise ValueError("No matching students found in roster.")
    return roster

def add_to_roster(conf, roster, name, username, section, force=False):
    student = {
        "name": name,
        "username": username,
        "section": section
    }

    logger.debug("{}".format(roster))

    if not force and any([s['username'] == username for s in roster]):
      raise DuplicateUserError("Student already exists in roster!")

    try:
        student["id"] = Repo.get_user_id(
            username, conf.gitlab_host, conf.token
        )
    except RepoError:
        logger.warning(
            "Student {} does not have a Gitlab account.".format(name)
        )

    roster.append(student)




#!/usr/bin/env python3
import argparse
import importlib
import logging

from colorlog import ColoredFormatter
from requests.exceptions import HTTPError

from assigner.baserepo import StudentRepo
from assigner.config import config_context
from assigner.roster_util import get_filtered_roster

logger = logging.getLogger(__name__)

description = "An automated tool for assigning programming homework."

subcommands = [
    "new",
    "assign",
    "open",
    "get",
    "lock",
    "unlock",
    "archive",
    "unarchive",
    "protect",
    "unprotect",
    "status",
    "import",
    "canvas",
    "set",
    "roster",
    "init",
]

@config_context
def manage_users(conf, args, level):
    """Creates a folder for the assignment in the CWD (or <path>, if specified)
    and clones each students' repository into subfolders.
    """
    hw_name = args.name
    dry_run = args.dry_run

    if dry_run:
        raise NotImplementedError("'--dry-run' is not implemented")

    host = conf.gitlab_host
    namespace = conf.namespace
    token = conf.token
    semester = conf.semester

    roster = get_filtered_roster(conf.roster, args.section, args.student)

    count = 0
    for student in roster:
        username = student["username"]
        student_section = student["section"]
        if "id" not in student:
            logging.warning(
                "Student {} does not have a gitlab account.".format(username)
            )
            continue
        full_name = StudentRepo.name(semester, student_section,
                                     hw_name, username)

        try:
            repo = StudentRepo(host, namespace, full_name, token)
            repo.edit_member(student["id"], level)
            count += 1
        except HTTPError:
            raise

    print("Changed {} repositories.".format(count))


@config_context
def manage_repos(conf, args, action):
    """Performs an action (lambda) on all student repos
    """
    hw_name = args.name
    dry_run = args.dry_run

    host = conf.gitlab_host
    namespace = conf.namespace
    token = conf.token
    semester = conf.semester

    roster = get_filtered_roster(conf.roster, args.section, args.student)

    count = 0
    for student in roster:
        username = student["username"]
        student_section = student["section"]
        if "id" not in student:
            logging.warning(
                "Student {} does not have a gitlab account.".format(username)
            )
            continue
        full_name = StudentRepo.name(semester, student_section,
                                     hw_name, username)

        try:
            repo = StudentRepo(host, namespace, full_name, token)
            if not dry_run:
                action(repo)
            count += 1
        except HTTPError:
            raise

    print("Changed {} repositories.".format(count))


def configure_logging():
    root_logger = logging.getLogger()
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)

    # Create a colorized formatter
    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
        datefmt=None,
        reset=True,
        log_colors={
            "DEBUG":    "cyan",
            "INFO":     "green",
            "WARNING":  "yellow",
            "ERROR":    "red",
            "CRITICAL": "red,bg_white",
        },
        secondary_log_colors={},
        style="%"
    )

    # Add the formatter to the console handler, and the console
    # handler to the root logger.
    console.setFormatter(formatter)
    root_logger.addHandler(console)


def make_parser():
    """Construct and return a CLI argument parser.
    """
        
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("--config", default="_config.yml",
                        help="Path a config file")
    parser.add_argument("--tracebacks", action="store_true",
                        help="Show full tracebacks")
    parser.add_argument("--verbosity", default="INFO",
                        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
                        help="Desired log level")

    # If no arguments are provided, show the usage screen
    parser.set_defaults(run=lambda x: parser.print_usage())

    # Set up subcommands for each package
    subparsers = parser.add_subparsers(title="subcommands")

    for name in subcommands:
        module = importlib.import_module("assigner.commands." + name)
        subparser = subparsers.add_parser(name, help=module.help)
        module.setup_parser(subparser)

    # The "help" command shows the help screen
    help_parser = subparsers.add_parser("help",
                                        help="Show this help screen and exit")
    help_parser.set_defaults(run=lambda x: parser.print_help())

    return parser


def main():
    """Entry point
    """
    # Configure logging
    configure_logging()

    # Parse CLI args
    parser = make_parser()
    args = parser.parse_args()

    # Set logging verbosity
    logging.getLogger().setLevel(args.verbosity)
    logging.getLogger("requests").setLevel(logging.WARNING)

    # Do it
    try:
        args.run(args)
    except Exception as e:
        if args.tracebacks:
            raise e
        if isinstance(e, KeyError):
            logger.error(str(e) + " is missing")
        else:
            logger.error(str(e))
        raise SystemExit(1) from e


if __name__ == "__main__":
    main()



import logging
import jsonschema
import yaml

from collections import UserDict

class DuplicateUserError(Exception):
    pass

def config_context(func):
    def wrapper(cmdargs, *args):
        with Config(cmdargs.config) as conf:
            return func(conf, cmdargs, *args)
    return wrapper


class Config(UserDict):
    """Context manager for config; automatically saves changes"""

    CONFIG_SCHEMA = {
        "$schema": "http://json-schema.org/schema#",

        "type": "object",
        "properties": {
            # GitLab private token
            "token": {
                "type": "string",
            },

            # GitLab domain (https://git.gitlab.com)
            "gitlab-host": {
                "type": "string",
            },

            # GitLab Namespace name
            "namespace": {
                "type": "string",
            },

            # GitLab Namespace ID (we'd have to retrieve that)
            "namespace-id": {
                "type": "integer",
            },

            # Verbose name of the course (might be unnecessary)
            "course-name": {
                "type": "string",
            },

            # Current semester
            "semester": {
                "type": "string",
                "pattern": r"^\d{4}-(SP|FS|SS)$"
            },

            # Roster of students
            "roster": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {

                        # Their full name
                        "name": {
                            "type": "string"
                        },

                        # Section
                        "section": {
                            "type": "string"
                        },

                        # Their GitLab username (single sign on)
                        "username": {
                            "type": "string",
                            "pattern": "^[\w\.\-]+$",
                        },

                        # Their GitLab id (might be handy, but we'd have
                        # to fetch it and save it). Should save time in
                        # the long run instead of constantly querying
                        "id": {
                            "type": "integer",
                        },
                    },
                    "required": ["name", "username", "section"],
                    "additionalProperties": False,
                },
            },

            # Canvas API token
            "canvas-token": {
                "type": "string",
            },
            # Canvas domain
            "canvas-host": {
                "type": "string",
            }
        },
        "required": ["gitlab-host", "namespace", "token", "semester"],
        "additionalProperties": False,
    }

    def __init__(self, filename):
        super().__init__()
        self._filename = filename
        try:
            with open(filename) as f:
                self.data = yaml.safe_load(f)
                jsonschema.validate(self.data, self.__class__.CONFIG_SCHEMA)
        except FileNotFoundError:
            pass  # Just make an empty config; create on __exit__()
        except jsonschema.ValidationError as e:
            logging.warning("Your configuration is not valid: " + e.message)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        with open(self._filename, "w") as f:
            yaml.dump(self.data, f, indent=2, default_flow_style=False)

        return False  # propagate exceptions from the calling context

    def __getattr__(self, key):
        attr = getattr(super(Config, self), key, None)
        if attr:
            return attr
        # Keys contained dashes can be called using an underscore
        key = key.replace("_", "-")

        try:
            return self.data[key]
        except KeyError as e:
            if key == "roster":
                self.data[key] = []
                return self.data[key]
            raise e



import logging

from requests.exceptions import HTTPError

from assigner.baserepo import BaseRepo, Repo
from assigner.config import config_context

help = "Create a new base repo"

logger = logging.getLogger(__name__)


@config_context
def new(conf, args):
    """Creates a new base repository for an assignment so that you can add the
    instructions, sample code, etc.
    """
    hw_name = args.name
    dry_run = args.dry_run
    host = conf.gitlab_host
    namespace = conf.namespace
    token = conf.token

    if dry_run:
        url = Repo.build_url(host, namespace, hw_name)
        print("Created repo for {}:\n\t{}\n\t{}".format(hw_name, url, "(ssh url not available)"))
    else:
        try:
            repo = BaseRepo.new(hw_name, namespace, host, token)
            print("Created repo for {}:\n\t{}\n\t{}".format(hw_name, repo.url, repo.ssh_url))
        except HTTPError as e:
            if e.response.status_code == 400:
                logger.warning("Repository {} already exists!".format(hw_name))
            else:
                raise


def setup_parser(parser):
    parser.add_argument("name",
                           help="Name of the assignment.")
    parser.add_argument("--dry-run", action="store_true",
                           help="Don't actually do it.")
    parser.set_defaults(run=new)



import logging

from assigner.config import config_context


help = "Set configuration values"

logger = logging.getLogger(__name__)


@config_context
def set_conf(conf, args):
    """Sets <key> to <value> in the config.
    """
    conf[args.key] = args.value


def setup_parser(parser):
    parser.add_argument("key", help="Key to set")
    parser.add_argument("value", help="Value to set")
    parser.set_defaults(run=set_conf)



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='assigner',

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    version='0.1.0',

    description='Automatically assign programming homework to students on GitLab',
    long_description=long_description,

    url='https://github.com/redkyn/assigner',

    author='N. Jarus, M. Wisely, & T. Morrow',

    author_email='jarus@mst.edu',

    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Education',
        'Topic :: Education',
        'Topic :: Software Development :: Version Control :: Git',
        'Environment :: Console',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4', # TODO what is the minimum py3 version we support?
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #keywords='sample setuptools development',

    packages=find_packages(),

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    install_requires=[
        'GitPython>=1',
        'PyYAML>=3.11',
        'colorlog>=2.6,<3', # TODO can we move to v3?
        'jsonschema>=2.5',
        'requests==2.9.1', # TODO can we upgrade?
        'PTable>=0.9',
        'progressbar2>=3',
        ],

    python_requires='>=3',

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    #extras_require={
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    data_files=[
        ('share/assigner', ['TUTORIAL.md', '_config.example.yml'])
    ],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    entry_points={
        'console_scripts': [
            'assigner=assigner:main',
        ],
    },
)
